import random

import factory

from wbportfolio.models import DividendTransaction


class DividendTransactionsFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = DividendTransaction

    currency_fx_rate = 1.0
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")
    underlying_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")
    value_date = factory.Faker("date_object")
    ex_date = factory.Faker("date_object")
    record_date = factory.LazyAttribute(lambda o: o.ex_date)
    retrocession = 1.0
    shares = factory.LazyAttribute(lambda o: random.randint(10, 10000))
    price = factory.LazyAttribute(lambda o: random.randint(10, 10000))



# Generated by Django 4.1.10 on 2023-07-24 15:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0043_alter_instrumentlistthroughmodel_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="instrumentlist",
            name="identifier",
            field=models.CharField(blank=True, max_length=255, unique=True),
        ),
        migrations.AddField(
            model_name="instrumentlistthroughmodel",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.RenameIndex(
            model_name="instrumentprice",
            new_name="instrument_price_base_idx",
            old_name="instrument_price_idx",
        ),
        migrations.AddIndex(
            model_name="instrumentprice",
            index=models.Index(fields=["calculated", "instrument"], name="instrument_price_idx1"),
        ),
        migrations.AddIndex(
            model_name="instrumentprice",
            index=models.Index(fields=["instrument"], name="instrument_price_idx2"),
        ),
    ]



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_analyst, is_portfolio_manager

ASSETPOSITION_MENUITEM = MenuItem(
    label="Invested Asset Positions",
    endpoint="wbportfolio:assetpositiongroupby-list",
    permission=ItemPermission(permissions=["wbfdm.view_instrument"], method=is_analyst),
)

AGGREGATED_ASSETPOSITION_LIQUIDITY_MENUITEM = MenuItem(
    label="Asset Positions Liquidity Table",
    endpoint="wbportfolio:aggregatedassetpositionliquidity-list",
    permission=ItemPermission(permissions=["wbfdm.view_instrument"], method=is_portfolio_manager),
)



from datetime import timedelta

import pytest
from faker import Faker
from pandas.tseries.offsets import BDay

from wbportfolio.models import Trade

fake = Faker()


@pytest.mark.django_db
class TestProductGroupModel:
    def test_init(self, product_group):
        assert product_group.id is not None

    def test_get_total_fund_aum(
        self,
        portfolio_factory,
        product_factory,
        instrument_price_factory,
        trade_factory,
        product_group,
        currency_fx_rates_factory,
        weekday,
    ):
        p1 = product_factory.create(parent=product_group, inception_date=weekday)
        p2 = product_factory.create(parent=product_group, inception_date=weekday)
        portfolio1 = p1.portfolio
        portfolio2 = p2.portfolio
        shares1 = 1000
        v1 = instrument_price_factory.create(date=weekday, instrument=p1, outstanding_shares=shares1)
        c11 = trade_factory.create(
            portfolio=portfolio1,
            shares=shares1,
            value_date=weekday,
            underlying_instrument=p1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        fx_rate1 = p1.currency.fx_rates.get(date=v1.date)

        c21 = trade_factory.create(
            portfolio=portfolio2,
            value_date=weekday,
            underlying_instrument=p2,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        v2 = instrument_price_factory.create(instrument=p2, date=weekday, outstanding_shares=c21.shares)
        fx_rate2 = p2.currency.fx_rates.get(date=weekday)
        p1.update_last_valuation_date()
        p2.update_last_valuation_date()
        assert float(product_group.get_total_fund_aum(weekday)) == pytest.approx(
            float(v1.net_value) * float(c11.shares) / float(fx_rate1.value)
            + float(v2.net_value) * float(c21.shares) / float(fx_rate2.value)
        )
        # test: get only trade at value_date
        trade_factory.create(portfolio=portfolio1, transaction_date=weekday, value_date=weekday + timedelta(days=1))
        assert float(product_group.get_total_fund_aum(weekday)) == pytest.approx(
            float(v1.net_value) * float(c11.shares) / float(fx_rate1.value)
            + float(v2.net_value) * float(c21.shares) / float(fx_rate2.value)
        )

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_fund_product_table(self, product_group, product_factory, instrument_price_factory, val_date):
        val_date = (val_date + BDay(1)).date()
        p1 = product_factory.create(parent=product_group, inception_date=val_date)
        p2 = product_factory.create(parent=product_group, inception_date=val_date)

        v1 = instrument_price_factory.create(instrument=p1, date=val_date, calculated=False)
        v2 = instrument_price_factory.create(instrument=p2, date=val_date, calculated=False)
        df = product_group.get_fund_product_table(v2.date)
        dfp1 = df[df["share class"] == p1.name]
        assert dfp1["share class"].iloc[0] == p1.name
        assert dfp1["launch date"].iloc[0] == f"{v1.date:%Y-%m-%d}"
        assert dfp1["price"].iloc[0] == f"{v1.net_value:,.2f}"


@pytest.mark.django_db
class TestProductGroupRepresentantModel:
    def test_init(self, product_group_representant):
        assert product_group_representant.id is not None



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AdjustmentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:adjustment-list", args=[], request=self.request)

    def get_delete_endpoint(self, **kwargs):
        return None


class AdjustmentEquityEndpointConfig(AdjustmentEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-adjustment-list", args=[self.view.kwargs["instrument_id"]], request=self.request
        )



from .composite import CompositeRebalancing
from .model_portfolio import ModelPortfolioRebalancing
from .equally_weighted import EquallyWeightedRebalancing
from .market_capitalization_weighted import MarketCapitalizationRebalancing



from decimal import Decimal

from django.db import models


class TransactionMixin(models.Model):
    value_date = models.DateField(
        verbose_name="Value Date",
        help_text="The date that this transaction was valuated/paid.",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="%(class)ss", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    underlying_instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        related_name="%(class)ss",
        limit_choices_to=models.Q(children__isnull=True),
        on_delete=models.PROTECT,
        verbose_name="Underlying Instrument",
        help_text="The instrument that is this transaction.",
    )
    currency = models.ForeignKey(
        "currency.Currency",
        related_name="%(class)ss",
        on_delete=models.PROTECT,
        verbose_name="Currency",
    )
    currency_fx_rate = models.DecimalField(
        max_digits=14, decimal_places=8, default=Decimal(1.0), verbose_name="FOREX rate"
    )
    price = models.DecimalField(
        max_digits=16,
        decimal_places=4,
        help_text="The price per share.",
        verbose_name="Price",
    )
    price_gross = models.DecimalField(
        max_digits=16,
        decimal_places=4,
        help_text="The gross price per share.",
        verbose_name="Gross Price",
    )
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal("0.0"),
        help_text="The number of shares held at record date, used to calculate the dividend",
        verbose_name="Shares / Quantity",
    )
    fees = models.GeneratedField(
        expression=models.F("price_gross") - models.F("price"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    total_value_gross = models.GeneratedField(
        expression=models.F("price_gross") * models.F("shares"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    total_value = models.GeneratedField(
        expression=models.F("price") * models.F("shares"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    total_value_fx_portfolio = models.GeneratedField(
        expression=models.F("currency_fx_rate") * models.F("price") * models.F("shares"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )
    total_value_gross_fx_portfolio = models.GeneratedField(
        expression=models.F("currency_fx_rate") * models.F("price_gross") * models.F("shares"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )

    comment = models.TextField(default="", verbose_name="Comment", blank=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        if not getattr(self, "currency", None) and self.underlying_instrument:
            self.currency = self.underlying_instrument.currency
        if self.currency_fx_rate is None:
            self.currency_fx_rate = self.underlying_instrument.currency.convert(
                self.value_date, self.portfolio.currency, exact_lookup=True
            )
        if self.price is not None and self.price_gross is None:
            self.price_gross = self.price
        elif self.price_gross is not None and self.price is None:
            self.price = self.price_gross
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



# Import CRM relevant data
from .roles import PortfolioRole
from .adjustments import Adjustment
from .asset import AssetPosition, AssetPositionGroupBy
from .custodians import Custodian
from .products import Product, FeeProductPercentage
from .indexes import Index
from .product_groups import ProductGroup, ProductGroupRepresentant
from .portfolio_relationship import (
    PortfolioInstrumentPreferredClassificationThroughModel,
    InstrumentPortfolioThroughModel,
    PortfolioBankAccountThroughModel,
)
from .portfolio import Portfolio, PortfolioPortfolioThroughModel
from .portfolio_cash_targets import PortfolioCashTarget
from .portfolio_cash_flow import DailyPortfolioCashFlow
from .portfolio_swing_pricings import PortfolioSwingPricing
from .registers import Register
from .transactions import *
from .reconciliations import AccountReconciliation, AccountReconciliationLine
from .signals import *



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_portfolio_manager

PORTFOLIO_DAILY_CASH_FLOW = MenuItem(
    label="Daily Portfolio Cash Flow",
    endpoint="wbportfolio:portfoliocashflow-list",
    permission=ItemPermission(permissions=["wbportfolio.view_portfolio"], method=is_portfolio_manager),
)



import random
from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import patch

import pandas as pd
import pytest
from django.contrib.contenttypes.models import ContentType
from django.db.models import F, Sum
from django.forms.models import model_to_dict
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import DateRange
from wbcore.contrib.geography.factories import CountryFactory

from wbportfolio.models import (
    AssetPosition,
    Portfolio,
    PortfolioInstrumentPreferredClassificationThroughModel,
    PortfolioPortfolioThroughModel,
    Trade,
)
from wbportfolio.models.asset import AssetPositionIterator

from ...models.portfolio import get_returns, update_portfolio_after_investable_universe
from .utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestPortfolioModel(PortfolioTestMixin):
    def test_init(self, portfolio):
        assert portfolio.id is not None

    def test_str(self, portfolio):
        assert str(portfolio) == f"{portfolio.id:06}: {portfolio.name}"

    def test_get_assets(self, portfolio, product, cash, asset_position_factory):
        asset_position_factory.create_batch(4, portfolio=portfolio, underlying_instrument=product)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=cash)
        assert portfolio._get_assets().count() == 5
        assert portfolio._get_assets(with_cash=False).count() == 4

    def test_get_earliest_asset_position_date(self, portfolio, asset_position_factory):
        asset_position_factory.create_batch(5, portfolio=portfolio)
        assert portfolio.get_earliest_asset_position_date() == AssetPosition.objects.earliest("date").date

    def test_get_latest_asset_position_date(self, portfolio, asset_position_factory):
        asset_position_factory.create_batch(5, portfolio=portfolio)
        assert portfolio.get_latest_asset_position_date() == AssetPosition.objects.latest("date").date

    def test_get_holding(self, portfolio_factory, asset_position_factory, equity, weekday):
        portfolio = portfolio_factory.create()
        asset_position_factory.create(portfolio=portfolio, date=weekday, initial_price=1, initial_shares=10)
        a2 = asset_position_factory.create(
            portfolio=portfolio,
            date=weekday,
            initial_price=1,
            initial_shares=40,
            underlying_instrument=equity,
            portfolio_created=portfolio_factory.create(),
        )
        asset_position_factory.create(portfolio=portfolio, date=weekday, initial_price=1, initial_shares=50)
        a4 = asset_position_factory.create(
            portfolio=portfolio,
            date=weekday,
            initial_price=1,
            initial_shares=30,
            underlying_instrument=equity,
            portfolio_created=portfolio_factory.create(),
        )
        assert (
            portfolio.get_holding(weekday)
            .filter(underlying_instrument=equity)
            .values_list("total_value_fx_portfolio", flat=True)[0]
            == a2._total_value_fx_portfolio + a4._total_value_fx_portfolio
        )

    def test_get_groupeby(self, portfolio, asset_position_factory, weekday):
        a1 = asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=10, weighting=0.1
        )
        asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=40, weighting=0.4
        )
        asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=50, weighting=0.5
        )

        def groupby(qs, **kwargs):
            return qs.annotate(aggregated_title=F("underlying_instrument__ticker"))

        df = portfolio._get_groupedby_df(groupby, weekday)
        assert df.aggregated_title[2] == a1.underlying_instrument.ticker

    def test_get_geographical_breakdown(self, portfolio, asset_position_factory, equity_factory, weekday):
        c1 = CountryFactory.create()
        c2 = CountryFactory.create()
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c1), date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c1), date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c2), date=weekday
        )
        assert portfolio.get_geographical_breakdown(weekday).shape[0] == 2

    def test_get_currency_exposure(self, portfolio, asset_position_factory, currency_factory, equity_factory, weekday):
        a1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=equity_factory.create(),
            currency=currency_factory.create(),
            date=weekday,
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(), currency=a1.currency, date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=equity_factory.create(),
            currency=currency_factory.create(),
            date=weekday,
        )
        assert portfolio.get_currency_exposure(weekday).shape[0] == 2

    def test_get_equity_market_cap_distribution(
        self, portfolio, equity, asset_position_factory, instrument_price_factory
    ):
        price = instrument_price_factory.create(instrument=equity)
        asset_position_factory.create_batch(10, portfolio=portfolio, date=price.date, underlying_instrument=equity)

        assert not portfolio.get_equity_market_cap_distribution(price.date).empty

    def test_get_get_equity_liquidity(self, portfolio, asset_position_factory, equity_factory, weekday):
        asset_position_factory.create_batch(
            10, portfolio=portfolio, date=weekday, underlying_instrument=equity_factory.create()
        )
        assert not portfolio.get_equity_liquidity(weekday).empty

    def test_get_industry_exposure(
        self, portfolio, asset_position_factory, weekday, classification_group, classification_factory, equity_factory
    ):
        parent_classification = classification_factory.create(group=classification_group)
        asset_position_factory.create_batch(
            10,
            portfolio=portfolio,
            date=weekday,
            underlying_instrument=equity_factory.create(
                classifications=[
                    classification_factory.create(group=classification_group, parent=parent_classification)
                ]
            ),
        )
        assert not portfolio.get_industry_exposure(weekday).empty

    def test_get_asset_allocation(self, portfolio, equity, cash, index_factory, asset_position_factory, weekday):
        index = index_factory.create(is_cash=True)
        asset_position_factory.create_batch(10, portfolio=portfolio, date=weekday, underlying_instrument=equity)
        asset_position_factory.create(portfolio=portfolio, date=weekday, underlying_instrument=index)
        asset_position_factory.create(portfolio=portfolio, date=weekday, underlying_instrument=cash)
        assert portfolio.get_asset_allocation(weekday).shape[0] == 2

    def test_get_portfolio_contribution_df(self, portfolio, asset_position_factory, instrument_factory, weekday):
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        end = (weekday + BDay(1)).date()
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            date=weekday,
            initial_price=100,
            initial_shares=10,
            initial_currency_fx_rate=1,
            weighting=0.25,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            date=weekday,
            initial_price=100,
            initial_shares=30,
            initial_currency_fx_rate=1,
            weighting=0.75,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            date=end,
            initial_price=120,
            initial_shares=10,
            initial_currency_fx_rate=1,
            weighting=0.33,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            date=end,
            initial_price=80,
            initial_shares=30,
            initial_currency_fx_rate=1,
            weighting=0.66,
        )
        res = portfolio.get_portfolio_contribution_df(weekday, end)
        assert 0.05 == pytest.approx(res.contribution_total[0])
        assert -0.15 == pytest.approx(res.contribution_total[1])
        assert 0.2 == pytest.approx(res.performance_total[0])
        assert -0.2 == pytest.approx(res.performance_total[1])

    def test_get_longshort_distribution(
        self, asset_position_factory, portfolio_factory, cash, index_factory, equity_factory, weekday
    ):
        portfolio = portfolio_factory.create()

        ind1 = index_factory.create(is_cash=False)
        short_underlying_portfolio = portfolio_factory.create()
        ind1.portfolios.add(short_underlying_portfolio)

        ind2 = index_factory.create(is_cash=False)
        long_underlying_portfolio = portfolio_factory.create()
        ind2.portfolios.add(long_underlying_portfolio)

        asset_position_factory.create(date=weekday, portfolio=portfolio, weighting=-1.0, underlying_instrument=ind1)
        short_p1 = asset_position_factory.create(
            date=weekday, portfolio=short_underlying_portfolio, underlying_instrument=cash
        )
        short_p2 = asset_position_factory.create(
            date=weekday,
            portfolio=short_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
            weighting=1 - short_p1.weighting,
        )

        asset_position_factory.create(date=weekday, portfolio=portfolio, weighting=1.0, underlying_instrument=ind2)
        long_p1 = asset_position_factory.create(
            date=weekday,
            portfolio=long_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
        )
        long_p2 = asset_position_factory.create(
            date=weekday,
            portfolio=long_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
            weighting=1 - long_p1.weighting,
        )

        res = portfolio.get_longshort_distribution(weekday)
        total_weight = abs(short_p2.weighting) + long_p1.weighting + long_p2.weighting
        assert Decimal(res.weighting[0]) == pytest.approx(
            (long_p1.weighting + long_p2.weighting) / total_weight, rel=Decimal(1e-4)
        )
        assert Decimal(res.weighting[1]) == pytest.approx((abs(short_p2.weighting)) / total_weight, rel=Decimal(1e-4))

    @patch.object(Portfolio, "estimate_net_asset_values", autospec=True)
    def test_change_at_date(self, mock_estimate_net_asset_values, asset_position_factory, portfolio, weekday):
        asset_position_factory.create_batch(10, portfolio=portfolio, date=weekday)

        portfolio.change_at_date(weekday, recompute_weighting=True)

        # test that change at date normalize the weighting
        total_value = AssetPosition.objects.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
        for pos in AssetPosition.objects.all():
            assert float(pos.weighting) == pytest.approx(float(pos.total_value_fx_portfolio / total_value), rel=1e-2)

        mock_estimate_net_asset_values.assert_called_once_with(portfolio, (weekday + BDay(1)).date(), weights=None)

    @patch.object(Portfolio, "compute_lookthrough", autospec=True)
    def test_change_at_date_with_dependent_portfolio(
        self,
        mock_compute_lookthrough,
        portfolio_factory,
        product_factory,
        instrument_price_factory,
        customer_trade_factory,
        weekday,
    ):
        base_portfolio = portfolio_factory.create()

        dependent_portfolio = portfolio_factory.create(is_lookthrough=True)
        dependent_portfolio.depends_on.add(base_portfolio)
        base_portfolio.change_at_date(weekday)

        mock_compute_lookthrough.assert_called_once_with(dependent_portfolio, weekday)

    def test_is_active_at_date(
        self,
        portfolio,
    ):
        # a portfolio is active at a date if it is active or the deletion time is greater than that date AND if there is instruments attached, at least one instrument is still active as well

        assert portfolio.is_active
        assert portfolio.is_active_at_date(fake.date_object())

        portfolio.delete()  # soft deletion
        assert portfolio.is_active_at_date(fake.past_date())
        assert not portfolio.is_active_at_date(fake.future_date())

    def test_is_active_at_date_with_instruments(
        self,
        portfolio,
        instrument_factory,
    ):
        i1 = instrument_factory.create(inception_date=date.today(), delisted_date=None)
        i2 = instrument_factory.create(inception_date=date.today(), delisted_date=None)
        portfolio.instruments.add(i1)
        portfolio.instruments.add(i2)

        assert i1.is_active_at_date(fake.future_date())
        assert i2.is_active_at_date(fake.future_date())
        assert portfolio.is_active_at_date(fake.future_date())

        i1.delisted_date = date.today()
        i1.save()
        assert portfolio.is_active_at_date(fake.future_date())

        i2.delisted_date = date.today()
        i2.save()
        assert not portfolio.is_active_at_date(
            fake.date_object()
        )  # as no instrument is active, even if the portfolio is active at any date, the portfolio is consiodered inactive

    def test_propagate_or_update_assets(
        self, portfolio, asset_position_factory, instrument_factory, instrument_price_factory, weekday
    ):
        next_day = (weekday + BDay(1)).date()

        i1 = instrument_factory.create(currency=portfolio.currency)
        price1_0 = instrument_price_factory.create(instrument=i1, date=(weekday - BDay(1)).date())  # noqa
        price1_1 = instrument_price_factory.create(instrument=i1, date=weekday)
        price1_2 = instrument_price_factory.create(instrument=i1, date=next_day)
        i2 = instrument_factory.create(currency=portfolio.currency)
        price2_0 = instrument_price_factory.create(instrument=i2, date=(weekday - BDay(1)).date())  # noqa
        price2_1 = instrument_price_factory.create(instrument=i2, date=weekday)
        price2_2 = instrument_price_factory.create(instrument=i2, date=next_day)
        i3 = instrument_factory.create(currency=portfolio.currency)
        price3_0 = instrument_price_factory.create(instrument=i3, date=(weekday - BDay(1)).date())  # noqa
        price3_1 = instrument_price_factory.create(instrument=i3, date=weekday)
        price3_2 = instrument_price_factory.create(instrument=i3, date=next_day)
        i4 = instrument_factory.create(currency=portfolio.currency)

        a1_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            underlying_quote_price=price1_1,
            date=weekday,
            weighting=Decimal(0.4),
        )
        a2_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            underlying_quote_price=price2_1,
            date=weekday,
            weighting=Decimal(0.3),
        )
        a3_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i3,
            underlying_quote_price=price3_1,
            date=weekday,
            weighting=Decimal(0.2),
        )
        a4_1 = asset_position_factory.create(  # noqa
            portfolio=portfolio,
            underlying_instrument=i4,
            underlying_quote_price=None,  # the price won't be created automatically by the fixture, we expect this position to be removed from the propagated portfolio
            date=weekday,
            weighting=Decimal(0.1),
        )

        # Test basic output
        portfolio.propagate_or_update_assets(weekday, next_day)
        a1_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i1)
        a2_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i2)
        a3_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i3)
        with pytest.raises(AssetPosition.DoesNotExist):
            AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i4)

        assert a1_2.initial_price == pytest.approx(price1_2.net_value, rel=Decimal(1e-4))
        assert a2_2.initial_price == pytest.approx(price2_2.net_value, rel=Decimal(1e-4))
        assert a3_2.initial_price == pytest.approx(price3_2.net_value, rel=Decimal(1e-4))

        contrib_1 = a1_1.weighting * price1_2.net_value / price1_1.net_value
        contrib_2 = a2_1.weighting * price2_2.net_value / price2_1.net_value
        contrib_3 = a3_1.weighting * price3_2.net_value / price3_1.net_value
        assert a1_2.weighting == pytest.approx(contrib_1 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))
        assert a2_2.weighting == pytest.approx(contrib_2 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))
        assert a3_2.weighting == pytest.approx(contrib_3 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))

        # # Test if a deleted assets is kept if delete_existing_assets is set to True
        # a1_1.delete()
        # portfolio.propagate_or_update_assets(weekday, next_day, delete_existing_assets=True)
        # with pytest.raises(AssetPosition.DoesNotExist):
        #     a1_2.refresh_from_db()

        # a2_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i2)
        # a3_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i3)
        # assert a2_2
        # assert a3_2

        # Test that we don't do anything on target portfolio because there is a non estimated position
        a2_2.is_estimated = False
        a2_2.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a2_2_weighting = a2_2.weighting
        a3_2_weighting = a3_2.weighting
        a2_2.refresh_from_db()
        a3_2.refresh_from_db()

        assert a2_2.weighting == a2_2_weighting
        assert a3_2.weighting == a3_2_weighting

    def test_propagate_or_update_assets_active_states(
        self, weekday, active_product, asset_position_factory, instrument_price_factory, instrument
    ):
        next_day = (weekday + BDay(1)).date()

        portfolio = active_product.portfolio
        portfolio.currency = instrument.currency
        portfolio.save()

        instrument_price_factory.create(date=(weekday - BDay(1)).date(), instrument=instrument)
        a1 = asset_position_factory.create(
            portfolio=portfolio, date=weekday, underlying_instrument=instrument, currency=instrument.currency
        )
        instrument_price_factory.create(date=next_day, instrument=instrument)

        active_product.delisted_date = weekday
        active_product.save()
        # Test1: test if unactive portfolio keep having the to date assets. (asset found at next day are suppose to be deleted when the portfolio is non active at the from date)
        portfolio.propagate_or_update_assets(weekday, next_day)
        assert portfolio.assets.filter(date=next_day).exists() is False

        # Activate product
        active_product.delisted_date = None
        active_product.save()

        # Expect proper creation
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future = AssetPosition.objects.get(portfolio=portfolio, date=next_day)

        # Test if only estimated update existing pos,

        portfolio.propagate_or_update_assets(weekday, next_day)
        initial_shares = a1.initial_shares

        # Test that estimated shares keep being updated
        portfolio.only_weighting = False
        portfolio.save()
        a1.initial_shares *= 2
        a1.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future = AssetPosition.objects.get(portfolio=portfolio, date=next_day)
        assert a_future.initial_shares == initial_shares * 2

        # Test that non-estimated shares are not being updated
        updated_fields = ["initial_currency_fx_rate", "weighting", "initial_price", "initial_shares"]
        for field in updated_fields:
            setattr(a1, field, getattr(a1, field) * 2)
        a1.save()
        a_future_copy = model_to_dict(a_future)
        a_future.is_estimated = False
        a_future.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future.refresh_from_db()
        for field in updated_fields:
            assert getattr(a_future, field) == a_future_copy[field]

        # Test active (from) portfolio but not active (to) create a zero weight position
        active_product.delisted_date = next_day
        active_product.save()

        a_future.is_estimated = True
        a_future.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        with pytest.raises(AssetPosition.DoesNotExist):
            AssetPosition.objects.get(portfolio=portfolio, date=next_day)
        # assert a_future.weighting == 0

    def test_update_preferred_classification_per_instrument(
        self, portfolio, asset_position_factory, equity_factory, classification_factory, classification_group_factory
    ):
        primary_group = classification_group_factory.create(is_primary=True)
        other_group = classification_group_factory.create(is_primary=False)
        c1 = classification_factory.create(group=other_group)
        c2_primary = classification_factory.create(group=primary_group)
        c2_secondary = classification_factory.create(group=other_group)
        c3_1 = classification_factory.create(group=other_group)
        c3_2 = classification_factory.create(group=other_group)

        # One classification to this instrument, we expect the relationship to be filled automatically
        i1 = equity_factory.create(classifications=[c1])
        # One classification "Primary" and one "other" to this instrument, we expect the relationship to be filled automatically
        i2 = equity_factory.create(classifications=[c2_secondary, c2_primary])
        # Two non-primary classifications to this instrument, we expect the relationship to not be filled with the classification automatically (created though)
        i3 = equity_factory.create(classifications=[c3_2, c3_1])
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=i1)
        a2 = asset_position_factory.create(portfolio=portfolio, underlying_instrument=i2)
        a3 = asset_position_factory.create(portfolio=portfolio, underlying_instrument=i3)

        assert not portfolio.preferred_instrument_classifications.exists()
        portfolio.update_preferred_classification_per_instrument()
        res1 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=c1, instrument=i1, classification_group=other_group
        )
        res2 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=c2_secondary, instrument=i2, classification_group=other_group
        )
        res3 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=None, instrument=i3, classification_group=None
        )

        assert not PortfolioInstrumentPreferredClassificationThroughModel.objects.exclude(
            id__in=[res1.id, res2.id, res3.id]
        ).exists()

        # We delete portfolio positions and retrigger the function to check that the leftovers relationship are indeed removed
        a3.delete()
        a2.delete()
        portfolio.update_preferred_classification_per_instrument()
        with pytest.raises(PortfolioInstrumentPreferredClassificationThroughModel.DoesNotExist):
            res3.refresh_from_db()
        with pytest.raises(PortfolioInstrumentPreferredClassificationThroughModel.DoesNotExist):
            res2.refresh_from_db()
        res1.refresh_from_db()
        assert res1

    def test_get_total_asset_under_management(
        self, portfolio, customer_trade_factory, instrument_factory, instrument_price_factory, weekday
    ):
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        previous_day = (weekday - BDay(5)).date()
        price11 = instrument_price_factory.create(instrument=i1, date=weekday, calculated=False)
        price12 = instrument_price_factory.create(instrument=i1, date=previous_day, calculated=False)
        price2 = instrument_price_factory.create(instrument=i2, date=weekday, calculated=False)

        # "noise" trades
        pending_customer_trade_i1 = customer_trade_factory.create(  # noqa
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1, pending=True
        )
        marked_for_deletion_customer_trade_i1 = customer_trade_factory.create(  # noqa
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1, marked_for_deletion=True
        )

        # valid trade for two different instrument but within the same portfolio
        trade_11 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1
        )
        trade_12 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=previous_day, underlying_instrument=i1
        )
        trade_2 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i2
        )

        assert (
            portfolio.get_total_asset_under_management(weekday)
            == price11.net_value * (trade_11.shares + trade_12.shares) + price2.net_value * trade_2.shares
        )
        assert portfolio.get_total_asset_under_management(previous_day) == price12.net_value * trade_12.shares
        assert portfolio.get_total_asset_under_management(previous_day - BDay(1)) == Decimal(0)

    def test_tracked_object(self, portfolio, asset_position_factory):
        assert not Portfolio.tracked_objects.exists()

        asset_position_factory.create(portfolio=portfolio)
        assert set(Portfolio.tracked_objects.all()) == {
            portfolio,
        }

        portfolio.is_tracked = False
        portfolio.save()
        assert portfolio.is_manageable is True
        assert Portfolio.tracked_objects.exists()

        portfolio.is_manageable = False
        portfolio.save()
        assert not Portfolio.tracked_objects.exists()

    def test_is_invested_at_date(self, portfolio_factory):
        portfolio = portfolio_factory.create(invested_timespan=DateRange(date(2024, 1, 2), date(2024, 1, 3)))
        assert portfolio.is_invested_at_date(date(2024, 1, 1)) is False
        assert portfolio.is_invested_at_date(date(2024, 1, 2)) is True
        assert portfolio.is_invested_at_date(date(2024, 1, 3)) is False

        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 1))) == set()
        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 2))) == {portfolio}
        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 1))) == set()

    @patch.object(Portfolio, "get_total_asset_under_management", autospec=True)
    def test_compute_lookthrough(
        self,
        mock_fct,
        active_product,
        weekday,
        portfolio_factory,
        index_factory,
        equity_factory,
        asset_position_factory,
        trade_factory,
        instrument_price_factory,
        instrument_portfolio_through_model_factory,
    ):
        root_index = index_factory.create()
        root_index_portfolio = portfolio_factory.create()
        root_index.portfolios.add(root_index_portfolio)

        index1 = index_factory.create()
        index1_portfolio = portfolio_factory.create()
        index1.portfolios.add(index1_portfolio)

        a1 = asset_position_factory.create(
            underlying_instrument=index1,
            portfolio=root_index_portfolio,
            weighting=0.6,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        index2 = index_factory.create()
        index2_portfolio = portfolio_factory.create()
        index2.portfolios.add(index2_portfolio)

        a2 = asset_position_factory.create(
            underlying_instrument=index2,
            portfolio=root_index_portfolio,
            weighting=0.4,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        a1_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.2,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a2_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.3,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a3_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.5,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        a1_2 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index2_portfolio,
            weighting=0.7,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a2_2 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index2_portfolio,
            weighting=0.3,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        product_base_portfolio = active_product.primary_portfolio
        product_portfolio = portfolio_factory.create(is_lookthrough=True, only_weighting=True)
        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=product_portfolio)
        trade_factory.create(
            underlying_instrument=active_product,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        product_portfolio.depends_on.add(root_index_portfolio)

        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=product_portfolio)

        instrument_price_factory.create(instrument=active_product, date=weekday)
        trade_factory.create(
            underlying_instrument=active_product,
            portfolio=product_base_portfolio,
            transaction_date=weekday,
            shares=1000,
            transaction_subtype="SUBSCRIPTION",
        )

        product_portfolio.compute_lookthrough(weekday)
        assert product_portfolio.assets.filter(date=weekday).count() == 5
        assert float(a1_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a1_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a2_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a2_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a3_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a3_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a1_2.weighting) * float(a2.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index2_portfolio,
                    underlying_instrument=a1_2.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a2_2.weighting) * float(a2.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index2_portfolio,
                    underlying_instrument=a2_2.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert Decimal(1.0) == pytest.approx(product_portfolio.assets.aggregate(s=Sum("weighting"))["s"])

        product_portfolio.only_weighting = False
        product_portfolio.save()
        mock_fct.return_value = Decimal(1_000_000)
        product_portfolio.compute_lookthrough(weekday)
        position = product_portfolio.assets.get(
            portfolio_created=index1_portfolio,
            underlying_instrument=a1_1.underlying_instrument,
            date=weekday,
        )
        assert position.initial_shares == (position.weighting * Decimal(1_000_000)) / (
            position.initial_price * position.initial_currency_fx_rate
        )

    def test_estimate_net_asset_values(
        self,
        weekday,
        equity_factory,
        product_factory,
        asset_position_factory,
        instrument_price_factory,
        trade_factory,
    ):
        while weekday.weekday() in [5, 6]:
            weekday += timedelta(days=1)

        previous_sync_date = weekday - timedelta(days=1)
        while previous_sync_date.weekday() in [5, 6]:
            previous_sync_date -= timedelta(days=1)

        product = product_factory.create(inception_date=weekday - timedelta(days=1), delisted_date=None)
        portfolio = product.portfolio

        trade_factory.create(
            underlying_instrument=product,
            portfolio=portfolio,
            transaction_date=previous_sync_date,
            shares=1000,
            transaction_subtype="SUBSCRIPTION",
        )

        e1 = equity_factory.create()
        pe1_1 = instrument_price_factory.create(instrument=e1, date=previous_sync_date)
        pe1_2 = instrument_price_factory.create(instrument=e1, date=weekday)
        e2 = equity_factory.create()
        pe2_1 = instrument_price_factory.create(instrument=e2, date=previous_sync_date)
        pe2_2 = instrument_price_factory.create(instrument=e2, date=weekday)
        e3 = equity_factory.create()
        pe3_1 = instrument_price_factory.create(instrument=e3, date=previous_sync_date)
        pe3_2 = instrument_price_factory.create(instrument=e3, date=weekday)

        a1_1 = asset_position_factory.create(
            underlying_instrument=e1,
            underlying_quote_price=pe1_1,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.3),
            initial_price=Decimal(100),
            initial_shares=300,
        )
        a2_1 = asset_position_factory.create(
            underlying_instrument=e2,
            underlying_quote_price=pe2_1,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.5),
            initial_price=Decimal(100),
            initial_shares=500,
        )
        a3_1 = asset_position_factory.create(
            underlying_instrument=e3,
            underlying_quote_price=pe3_1,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.2),
            initial_price=Decimal(100),
            initial_shares=200,
        )

        a1_2 = asset_position_factory.create(
            underlying_instrument=e1,
            underlying_quote_price=pe1_2,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.3719),
            initial_price=Decimal(150),
            initial_shares=300,
        )
        a2_2 = asset_position_factory.create(
            underlying_instrument=e2,
            underlying_quote_price=pe2_2,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.4959),
            initial_price=Decimal(120),
            initial_shares=500,
        )
        a3_2 = asset_position_factory.create(
            underlying_instrument=e3,
            underlying_quote_price=pe3_2,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.1322),
            initial_price=Decimal(80),
            initial_shares=200,
        )

        price = instrument_price_factory.create(instrument=product, date=previous_sync_date, net_value=100)
        portfolio.estimate_net_asset_values(weekday)

        total_perf = (
            (a1_2._price / a1_1._price - 1) * a1_1.weighting
            + (a2_2._price / a2_1._price - 1) * a2_1.weighting
            + (a3_2._price / a3_1._price - 1) * a3_1.weighting
        )
        assert product.prices.count() == 2
        assert float(price.net_value * (Decimal(1.0) + total_perf)) == pytest.approx(
            float(product.prices.filter(date=weekday).first().net_value)
        )

    def test_pms_instrument(self, product_group, index, product, portfolio):
        product_group.portfolios.set([portfolio])
        product.portfolios.set([portfolio])
        index.portfolios.set([portfolio])
        assert set(portfolio.pms_instruments) == {product_group, product, index}

    @pytest.mark.parametrize(
        "portfolio__is_manageable, portfolio__is_lookthrough",
        [
            (True, True),
            (False, True),
            (False, False),
        ],
    )
    def test_cannot_be_rebalanced(self, portfolio):
        assert portfolio.can_be_rebalanced is False

    @pytest.mark.parametrize(
        "portfolio__is_manageable, portfolio__is_lookthrough",
        [
            (True, False),
        ],
    )
    def test_can_be_rebalanced(self, portfolio):
        assert portfolio.can_be_rebalanced is True

    def test_get_analytic_portfolio(
        self, weekday, portfolio, asset_position_factory, instrument_factory, instrument_price_factory
    ):
        i1 = instrument_factory.create(currency=portfolio.currency)
        i2 = instrument_factory.create(currency=portfolio.currency)
        p10 = instrument_price_factory.create(instrument=i1, date=weekday)
        p11 = instrument_price_factory.create(instrument=i1, date=(weekday + BDay(1)).date())
        p20 = instrument_price_factory.create(instrument=i2, date=weekday)
        p21 = instrument_price_factory.create(instrument=i2, date=(weekday + BDay(1)).date())

        a1 = asset_position_factory.create(date=weekday, portfolio=portfolio, underlying_instrument=i1)
        a1.refresh_from_db()
        a2 = asset_position_factory.create(
            date=weekday, portfolio=portfolio, underlying_instrument=i2, weighting=Decimal(1.0) - a1.weighting
        )
        a2.refresh_from_db()

        analytic_portfolio = portfolio.get_analytic_portfolio(weekday)
        assert analytic_portfolio.weights.tolist() == [float(a1.weighting), float(a2.weighting)]
        expected_X = pd.DataFrame(
            [[float(p11.net_value / p10.net_value - Decimal(1)), float(p21.net_value / p20.net_value - Decimal(1))]],
            columns=[i1.id, i2.id],
            index=[(weekday + BDay(1)).date()],
        )
        expected_X.index = pd.to_datetime(expected_X.index)
        pd.testing.assert_frame_equal(analytic_portfolio.X, expected_X, check_names=False, check_freq=False)

    def test_get_total_asset_value(self, weekday, portfolio, asset_position_factory):
        a1 = asset_position_factory.create(date=weekday, portfolio=portfolio)
        a2 = asset_position_factory.create(date=weekday, portfolio=portfolio)
        a3 = asset_position_factory.create(date=weekday, portfolio=portfolio)
        assert (
            portfolio.get_total_asset_value(weekday)
            == a1.initial_price * a1.initial_shares * a1.initial_currency_fx_rate
            + a2.initial_price * a2.initial_shares * a2.initial_currency_fx_rate
            + a3.initial_price * a3.initial_shares * a3.initial_currency_fx_rate
        )

    @pytest.mark.parametrize("name", fake.word())
    def test_create_model_portfolio(self, name, currency):
        portfolio = Portfolio.create_model_portfolio(name, currency)
        assert portfolio.name == name
        assert portfolio.currency == currency
        assert portfolio.is_manageable is True
        pms_instruments = list(portfolio.pms_instruments)
        assert len(pms_instruments) == 1
        assert pms_instruments[0].instrument_type.key == "index"
        assert pms_instruments[0].name == name
        assert pms_instruments[0].currency == currency

    @patch.object(Portfolio, "propagate_or_update_assets", autospec=True)
    def test_update_portfolio_after_investable_universe(
        self, mock_fct, weekday, portfolio_factory, asset_position_factory
    ):
        untracked_portfolio = portfolio_factory.create(is_tracked=False, is_manageable=False)  # noqa
        asset_position_factory.create(portfolio=untracked_portfolio)
        tracked_lookthrough_portfolio = portfolio_factory.create(is_tracked=True, is_lookthrough=True)  # noqa
        asset_position_factory.create(portfolio=tracked_lookthrough_portfolio)

        update_portfolio_after_investable_universe(end_date=weekday)
        mock_fct.assert_not_called()

        tracked_portfolio = portfolio_factory.create(is_tracked=True)
        asset_position_factory.create(portfolio=tracked_portfolio)

        update_portfolio_after_investable_universe(end_date=weekday)
        mock_fct.assert_called_once_with(tracked_portfolio, (weekday - BDay(1)).date(), weekday)

    def test_get_weights(
        self, weekday, portfolio_factory, asset_position_factory, instrument_factory, instrument_price_factory
    ):
        portfolio = portfolio_factory.create()
        portfolio_created = portfolio_factory.create()

        a1 = asset_position_factory.create(date=weekday, portfolio=portfolio)
        a2 = asset_position_factory.create(date=weekday, portfolio=portfolio)
        a3 = asset_position_factory.create(
            date=weekday,
            portfolio=portfolio,
            underlying_instrument=a2.underlying_instrument,
            portfolio_created=portfolio_created,
        )
        a1.refresh_from_db()
        a2.refresh_from_db()
        a3.refresh_from_db()
        weights = portfolio.get_weights(weekday)
        assert weights[a1.underlying_quote.id] == float(a1.weighting)
        assert weights[a2.underlying_quote.id] == float(a2.weighting + a3.weighting)

    def test_get_estimated_portfolio_from_weights(
        self, weekday, portfolio, instrument, instrument_price_factory, currency_fx_rates_factory
    ):
        p = instrument_price_factory.create(instrument=instrument, date=weekday)
        fx_portfolio = currency_fx_rates_factory.create(currency=portfolio.currency, date=weekday)
        fx_instrument = currency_fx_rates_factory.create(currency=instrument.currency, date=weekday)
        instrument_id: int = instrument.id
        weights = {instrument_id: random.random()}
        positions = AssetPositionIterator(
            portfolio, prices={weekday: {instrument_id: p.net_value}}, infer_underlying_quote_price=False
        )
        positions.add((weekday, weights))

        res = list(positions)
        a = res[0]
        assert len(res) == 1
        assert a.date == weekday
        assert a.underlying_quote == instrument
        assert a.underlying_quote_price == None
        assert a.initial_price == p.net_value
        assert a.weighting == weights[instrument.id]
        assert a.currency_fx_rate_portfolio_to_usd == fx_portfolio
        assert a.currency_fx_rate_instrument_to_usd == fx_instrument

        # ensure saving the unsave assetposition do not lead to exception
        a.save()
        assert a

    def test_drift_weights_with_rebalancer(
        self,
        weekday,
        rebalancer_factory,
        portfolio,
        asset_position_factory,
        instrument_price_factory,
        instrument_factory,
    ):
        middle_date = (weekday + BDay(1)).date()
        rebalancing_date = (middle_date + BDay(1)).date()

        i1 = instrument_factory.create(currency=portfolio.currency)
        i2 = instrument_factory.create(currency=portfolio.currency)
        asset_position_factory.create(date=weekday, portfolio=portfolio, underlying_instrument=i1, weighting=0.7)
        asset_position_factory.create(date=weekday, portfolio=portfolio, underlying_instrument=i2, weighting=0.3)

        instrument_price_factory.create(instrument=i1, date=(weekday - BDay(1)).date())
        instrument_price_factory.create(instrument=i1, date=weekday)
        instrument_price_factory.create(instrument=i1, date=middle_date)
        instrument_price_factory.create(instrument=i1, date=rebalancing_date)
        instrument_price_factory.create(instrument=i2, date=(weekday - BDay(1)).date())
        instrument_price_factory.create(instrument=i2, date=weekday)
        instrument_price_factory.create(instrument=i2, date=middle_date)
        instrument_price_factory.create(instrument=i2, date=rebalancing_date)

        rebalancer_factory.create(portfolio=portfolio, frequency="RRULE:FREQ=DAILY;", activation_date=rebalancing_date)
        positions, rebalancing_trade_proposal = portfolio.drift_weights(weekday, (rebalancing_date + BDay(1)).date())
        assert rebalancing_trade_proposal.trade_date == rebalancing_date
        assert rebalancing_trade_proposal.status == "SUBMIT"
        assert set(positions.get_weights().keys()) == {
            middle_date,
        }, "Drifting weight with a non automatic rebalancer stops the iteration"

        # we expect a equally rebalancing (default) so both trades needs to be created
        t1 = rebalancing_trade_proposal.trades.get(transaction_date=rebalancing_date, underlying_instrument=i1)
        t2 = rebalancing_trade_proposal.trades.get(transaction_date=rebalancing_date, underlying_instrument=i2)
        assert t1._target_weight == Decimal("0.5")
        assert t2._target_weight == Decimal("0.5")

        # we approve the rebalancing trade proposal
        assert rebalancing_trade_proposal.status == "SUBMIT"
        rebalancing_trade_proposal.approve(no_replay=True)
        rebalancing_trade_proposal.save()

        # check that the rebalancing was applied and position reflect that
        assert portfolio.assets.get(date=rebalancing_date, underlying_instrument=i1).weighting == Decimal("0.5")
        assert portfolio.assets.get(date=rebalancing_date, underlying_instrument=i2).weighting == Decimal("0.5")

    @patch("wbportfolio.models.portfolio.compute_metrics_as_task.delay")
    def test_bulk_create_positions(
        self, mock_compute_metrics, portfolio, weekday, asset_position_factory, instrument_factory
    ):
        portfolio.is_manageable = False
        portfolio.save()
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        i3 = instrument_factory.create()
        a1 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i1)

        # check initial creation
        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a1]))
        assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == a1.weighting
        assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).underlying_instrument == i1

        mock_compute_metrics.assert_called_once_with(
            weekday, basket_id=portfolio.id, basket_content_type_id=ContentType.objects.get_for_model(Portfolio).id
        )

        # check that if we change key value, an already exising position will be updated accordingly
        a1.weighting = Decimal(0.5)
        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a1]))
        assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == Decimal(0.5)

        a2 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i2)
        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a2]))
        assert (
            AssetPosition.objects.get(portfolio=portfolio, date=weekday, underlying_instrument=i1).weighting
            == a1.weighting
        )
        assert (
            AssetPosition.objects.get(portfolio=portfolio, date=weekday, underlying_instrument=i2).weighting
            == a2.weighting
        )

        a3 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i3)
        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a3]), delete_leftovers=True)
        assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == a3.weighting
        assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).underlying_instrument == i3

    def test_to_dependency_iterator(self, portfolio_factory, asset_position_factory, index, weekday):
        Portfolio.objects.all().delete()  # ensure no portfolio remains
        dependant_portfolio = portfolio_factory.create(name="dependant portfolio", id=1)
        dependency_portfolio = portfolio_factory.create(name="dependency portfolio", id=2)
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=dependant_portfolio, dependency_portfolio=dependency_portfolio
        )
        index_portfolio = portfolio_factory.create(name="Index portfolio", id=3)
        index.portfolios.add(index_portfolio)
        asset_position_factory.create(portfolio=dependency_portfolio, underlying_instrument=index, date=weekday)

        undependant_portfolio = portfolio_factory.create(name="undependant portfolio", id=4)
        res = list(Portfolio.objects.all().to_dependency_iterator(weekday))
        assert res == [index_portfolio, dependency_portfolio, dependant_portfolio, undependant_portfolio]

    def test_get_returns(self, instrument_factory, instrument_price_factory, asset_position_factory, portfolio):
        v1 = date(2025, 1, 1)
        v2 = date(2025, 1, 2)
        v3 = date(2025, 1, 3)

        i1 = instrument_factory.create()
        i2 = instrument_factory.create()

        i11 = instrument_price_factory.create(date=v1, instrument=i1)
        i12 = instrument_price_factory.create(date=v2, instrument=i1)
        i13 = instrument_price_factory.create(date=v3, instrument=i1)
        asset_position_factory.create(date=v1, portfolio=portfolio, underlying_instrument=i1)
        asset_position_factory.create(date=v3, portfolio=portfolio, underlying_instrument=i2)
        i11.refresh_from_db()
        i12.refresh_from_db()
        i13.refresh_from_db()
        returns = get_returns([i1.id, i2.id], from_date=v1, to_date=v3)

        expected_returns = pd.DataFrame(
            [[i12.net_value / i11.net_value - 1, 0.0], [i13.net_value / i12.net_value - 1, 0.0]],
            index=[v2, v3],
            columns=[i1.id, i2.id],
            dtype="float64",
        )
        expected_returns.index = pd.to_datetime(expected_returns.index)
        pd.testing.assert_frame_equal(returns, expected_returns, check_names=False, check_freq=False, atol=1e-6)

    @patch.object(Portfolio, "compute_lookthrough", autospec=True)
    def test_handle_controlling_portfolio_change_at_date(self, mock_compute_lookthrough, weekday, portfolio_factory):
        primary_portfolio = portfolio_factory.create(only_weighting=True)
        lookthrough_portfolio = portfolio_factory.create(is_lookthrough=True, only_weighting=False)
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=lookthrough_portfolio,
            dependency_portfolio=primary_portfolio,
            type=PortfolioPortfolioThroughModel.Type.PRIMARY,
        )

        primary_portfolio.handle_controlling_portfolio_change_at_date(weekday)
        mock_compute_lookthrough.assert_called_once_with(lookthrough_portfolio, weekday)



# Generated by Django 4.2.9 on 2024-02-02 10:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0050_fees_fee_date_fees_wbportfolio_transac_1f7a29_idx"),
    ]

    operations = []



# Generated by Django 5.0.6 on 2024-06-06 08:55

from decimal import Decimal

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0008_bankingcontact_access"),
        ("io", "0007_alter_exportsource_query_params"),
        ("wbportfolio", "0060_alter_portfolioportfoliothroughmodel_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="portfolio",
            name="bank_accounts",
            field=models.ManyToManyField(
                blank=True, related_name="wbportfolio_portfolios", to="directory.bankingcontact"
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="bank_account",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="wbportfolio_products",
                to="directory.bankingcontact",
            ),
        ),
        migrations.CreateModel(
            name="PortfolioCashTarget",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("valid_date", models.DateField()),
                ("min_target", models.DecimalField(blank=True, decimal_places=4, max_digits=5, null=True)),
                ("target", models.DecimalField(decimal_places=4, default=Decimal("0"), max_digits=5)),
                ("max_target", models.DecimalField(blank=True, decimal_places=4, max_digits=5, null=True)),
                ("comment", models.TextField(blank=True, default="")),
                (
                    "portfolio",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="cash_targets",
                        to="wbportfolio.portfolio",
                    ),
                ),
            ],
            options={
                "verbose_name": "Portfolio Cash Target",
                "verbose_name_plural": "Portfolio Cash Targets",
            },
        ),
        migrations.CreateModel(
            name="PortfolioSwingPricing",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("valid_date", models.DateField()),
                ("negative_threshold", models.DecimalField(decimal_places=4, max_digits=4)),
                ("negative_swing_factor", models.DecimalField(decimal_places=4, max_digits=4)),
                ("positive_threshold", models.DecimalField(decimal_places=4, max_digits=4)),
                ("positive_swing_factor", models.DecimalField(decimal_places=4, max_digits=4)),
                (
                    "portfolio",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="swing_pricings",
                        to="wbportfolio.portfolio",
                    ),
                ),
            ],
            options={
                "verbose_name": "Portfolio Swing Pricing",
                "verbose_name_plural": "Portfolio Swing Pricings",
            },
        ),
        migrations.CreateModel(
            name="DailyPortfolioCashFlow",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("pending", models.BooleanField(default=False)),
                ("value_date", models.DateField()),
                ("cash", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                (
                    "cash_flow_forecast",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=19),
                ),
                ("total_assets", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                (
                    "min_target_cash_pct",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=4),
                ),
                (
                    "target_cash_pct",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=4),
                ),
                (
                    "max_target_cash_pct",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=4),
                ),
                ("estimated_total_assets", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                ("cash_flow_asset_ratio", models.DecimalField(blank=True, decimal_places=4, max_digits=5)),
                ("true_cash", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                ("cash_pct", models.DecimalField(blank=True, decimal_places=4, max_digits=5)),
                ("true_cash_pct", models.DecimalField(blank=True, decimal_places=4, max_digits=5)),
                ("target_cash", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                ("excess_cash", models.DecimalField(blank=True, decimal_places=4, max_digits=19)),
                (
                    "proposed_rebalancing",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=19),
                ),
                (
                    "rebalancing",
                    models.DecimalField(blank=True, decimal_places=4, default=Decimal("0"), max_digits=19),
                ),
                ("comment", models.TextField(blank=True, default="")),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
                    ),
                ),
                (
                    "portfolio",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="daily_cashflows",
                        to="wbportfolio.portfolio",
                    ),
                ),
                (
                    "swing_pricing",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="cash_flow",
                        to="wbportfolio.portfolioswingpricing",
                    ),
                ),
            ],
            options={
                "verbose_name": "Daily Portfolio CashFlow",
                "verbose_name_plural": "Daily Portfolio CashFlow",
            },
        ),
        migrations.AddConstraint(
            model_name="portfoliocashtarget",
            constraint=models.UniqueConstraint(fields=("portfolio", "valid_date"), name="unique_portfolio_valid_date"),
        ),
        migrations.AddConstraint(
            model_name="portfolioswingpricing",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("negative_threshold__lt", 0),
                    ("negative_swing_factor__lt", 0),
                    ("positive_threshold__gt", 0),
                    ("positive_swing_factor__gt", 0),
                ),
                name="value_polarity",
            ),
        ),
        migrations.AddConstraint(
            model_name="portfolioswingpricing",
            constraint=models.UniqueConstraint(fields=("valid_date", "portfolio"), name="unique_valid_date_portfolio"),
        ),
        migrations.AddConstraint(
            model_name="dailyportfoliocashflow",
            constraint=models.UniqueConstraint(fields=("portfolio", "value_date"), name="unique_portfolio_value_date"),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbportfolio.models.transactions.claim import Claim


class ClaimEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.request.user.has_perm("wbportfolio.change_claim"):
            try:
                obj = self.view.get_object()
                if obj.status == Claim.Status.DRAFT:
                    return reverse("wbportfolio:claim-detail", args=[obj.id], request=self.request)
            except AssertionError:
                return reverse("wbportfolio:claim-list", request=self.request)
        return None

    def get_delete_endpoint(self, **kwargs):
        if self.request.user.has_perm("wbportfolio.change_claim"):
            try:
                obj = self.view.get_object()
                if obj.status == Claim.Status.WITHDRAWN:
                    return reverse("wbportfolio:claim-detail", args=[obj.id], request=self.request)
            except AssertionError:
                pass
        return None

    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:claim-list", request=self.request)


class ClaimProductEndpointConfig(ClaimEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:product-claim-list",
            args=[self.view.kwargs["product_id"]],
            request=self.request,
        )


class ClaimAccountEndpointConfig(ClaimEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:account-claim-list",
            args=[self.view.kwargs["account_id"]],
            request=self.request,
        )


class ClaimEntryEndpointConfig(ClaimEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:entry-claim-list",
            args=[self.view.kwargs["entry_id"]],
            request=self.request,
        )


class ClaimTradeEndpointConfig(ClaimEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:trade-claim-list",
            args=[self.view.kwargs["trade_id"]],
            request=self.request,
        )


class ConsolidatedTradeSummaryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class ConsolidatedTradeSummaryDistributionChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:consolidatedtradesummarydistributionchart-list", request=self.request)


class CumulativeNNMChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbportfolio:cumulativennmchart-list", request=self.request)


class ProfitAndLossPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class NegativeTermimalAccountPerProductEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return None

    def _get_instance_endpoint(self, **kwargs):
        return (
            reverse("wbportfolio:claim-list", request=self.request) + "?account={{account_id}}&product={{product_id}}"
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AssetAllocationModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if company_id := self.view.kwargs.get("company_id", None):
            return reverse(
                "company_portfolio:companyassetallocation-list",
                kwargs={"company_id": company_id},
                request=self.request,
            )
        return reverse("company_portfolio:assetallocation-list", request=self.request)

    def get_create_endpoint(self, **kwargs):
        if company_id := self.view.kwargs.get("company_id", None):
            return f"{reverse('company_portfolio:companyassetallocation-list', kwargs={'company_id': company_id}, request=self.request)}?company={company_id}"
        return super().get_create_endpoint(**kwargs)


class GeographicFocusModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if company_id := self.view.kwargs.get("company_id", None):
            return reverse(
                "company_portfolio:companygeographicfocus-list",
                kwargs={"company_id": company_id},
                request=self.request,
            )
        return reverse("company_portfolio:geographicfocus-list", request=self.request)

    def get_create_endpoint(self, **kwargs):
        if company_id := self.view.kwargs.get("company_id", None):
            return f"{reverse('company_portfolio:companygeographicfocus-list', kwargs={'company_id': company_id}, request=self.request)}?company={company_id}"
        return super().get_create_endpoint(**kwargs)



from contextlib import suppress
from datetime import date, timedelta

from django.db.models import Exists, IntegerField, Max, OuterRef, Q, QuerySet, Sum
from django.db.models.functions import Cast
from pandas.tseries.offsets import BDay, BQuarterEnd
from wbcore import filters as wb_filters
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.geography.models import Geography
from wbcore.pandas.filterset import PandasFilterSetMixin
from wbcore.utils.date import current_financial_month
from wbfdm.models import Classification, ClassificationGroup, Instrument, InstrumentType
from wbfdm.preferences import get_default_classification_group

from wbportfolio.models import AssetPosition, AssetPositionGroupBy, Portfolio, Product


class FilterByPositionMixin:
    def _get_products(self):
        return Product.objects.all()

    def _filter_by_positions(
        self,
        qs: QuerySet,
        underlying_instrument__key: str,
        current_positions: bool = False,
    ) -> QuerySet:
        asset_positions = AssetPosition.objects
        if current_positions:
            last_date = asset_positions.latest("date").date
            asset_positions = asset_positions.filter(date=last_date).exclude(weighting=0)
        products = self._get_products()
        return qs.filter(
            Exists(
                asset_positions.filter(
                    underlying_instrument_id=OuterRef(underlying_instrument__key),
                    portfolio__instruments__in=products,
                ).distinct()
            )
        )


def get_asset_filter_param(request, view):
    filter_params = {"is_investable": True}
    if portfolio_id := view.kwargs.get("portfolio_id", None):
        filter_params["portfolio"] = portfolio_id
    if date := request.GET.get("date", None):
        filter_params["date"] = date
    return filter_params


def get_latest_asset_position(field, request, view):
    qs = AssetPosition.objects
    with suppress(Instrument.DoesNotExist):
        instrument = Instrument.objects.get(
            id=view.kwargs.get("instrument_id", request.GET.get("underlying_instrument", None))
        )
        qs = qs.filter(underlying_instrument__in=instrument.get_descendants(include_self=True))
    if "portfolio_id" in view.kwargs:
        qs = qs.filter(portfolio__id=view.kwargs["portfolio_id"])
    if "portfolio" in request.GET:
        qs = qs.filter(portfolio__id=request.GET["portfolio"])
    if qs.exists():
        return qs.latest("date").date
    return date.today()


def get_latest_end_quarter_date_asset_position(field, request, view):
    val_date = get_latest_asset_position(field, request, view)
    if val_date:
        return (val_date - BQuarterEnd()).date()
    return None


def get_last_weekday(field, request, view):
    t = date.today() - timedelta(days=1)
    while t.weekday() > 4:
        t = t - timedelta(days=1)
    return t


def get_related_products(field, request, view):
    if "equity_id" in view.kwargs:
        return Product.objects.filter(equity_positions__equity__id=view.kwargs["equity_id"])
    return Product.objects.all()


def get_default_product_id(field, request, view):
    return get_related_products(field, request, view).first().id


def get_portfolio_filter_params(request, view):
    if underlying_instrument := request.GET.get("underlying_instrument", None):
        return {"instrument": underlying_instrument}
    return {}


def get_portfolio_default(field, request, view):
    portfolio = Portfolio.objects.get(id=view.kwargs.get("portfolio_id"))
    if portfolio.assets.exists():
        return portfolio.id
    return Portfolio.objects.first().id


def latest_portfolio_date_based_on_instrument(field, request, view):
    portfolio = Portfolio.objects.filter(id=get_portfolio_default(field, request, view))
    if portfolio.exists() and portfolio.first().assets.exists():
        return portfolio.first().assets.latest("date").date
    return date.today()


class DateFilterMixin:
    def start_filter(self, queryset, name, value):
        if value:
            if value.weekday() in [5, 6]:
                value = value - BDay(1)
            return queryset.filter(date__gte=value)
        return queryset

    def end_filter(self, queryset, name, value):
        return queryset.filter(date__lte=value)


class AssetPositionFilter(wb_filters.FilterSet):
    date = wb_filters.DateFilter(
        label="Date", lookup_expr="exact", field_name="date", initial=get_latest_asset_position, required=True
    )

    underlying_instrument__is_cash = wb_filters.BooleanFilter(
        label="With Cash", exclude=True, lookup_expr="exact", field_name="underlying_instrument__is_cash"
    )
    underlying_instrument__instrument_type = wb_filters.ModelChoiceFilter(
        label="Instrument Type",
        endpoint=InstrumentType.get_representation_endpoint(),
        value_key=InstrumentType.get_representation_value_key(),
        label_key=InstrumentType.get_representation_label_key(),
        queryset=InstrumentType.objects.all(),
        lookup_expr="exact",
        field_name="underlying_instrument__instrument_type",
    )

    classification = wb_filters.ModelChoiceFilter(
        label="Industry Classification",
        queryset=Classification.objects.all(),
        endpoint=Classification.get_representation_endpoint(),
        value_key=Classification.get_representation_value_key(),
        label_key=Classification.get_representation_label_key(),
        method="filter_classification",
        filter_params=get_asset_filter_param,
    )

    underlying_instrument__country = wb_filters.ModelChoiceFilter(
        label="Country",
        queryset=Geography.countries.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
        lookup_expr="exact",
        field_name="underlying_instrument__country",
    )

    portfolio_instrument = wb_filters.ModelChoiceFilter(
        label="Portfolio Instrument",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"is_managed": True},
        method="filter_portfolio_instrument",
    )

    underlying_instrument = wb_filters.ModelChoiceFilter(
        label="Underlying Instrument",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params=get_asset_filter_param,
    )
    hide_empty_position = wb_filters.BooleanFilter(
        initial=False, method="boolean_hide_empty_position", label="Hide Empty Position"
    )

    total_value_fx_usd = wb_filters.NumberFilter(label="Total Value ($)")
    total_value_fx_usd__gte = wb_filters.NumberFilter(
        lookup_expr="gte", field_name="total_value_fx_usd", label="Total Value ($)"
    )
    total_value_fx_usd__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="Total Value ($)",
        field_name="total_value_fx_usd",
    )

    price = wb_filters.NumberFilter(label="Price")
    price__gte = wb_filters.NumberFilter(lookup_expr="gte", field_name="price", label="Price")
    price__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="Price",
        field_name="price",
    )

    total_value_fx_portfolio = wb_filters.NumberFilter(label="Total Value (Portfolio)")
    total_value_fx_portfolio__gte = wb_filters.NumberFilter(
        lookup_expr="gte", field_name="total_value_fx_portfolio", label="Total Value (Portfolio)"
    )
    total_value_fx_portfolio__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="Total Value (Portfolio)",
        field_name="total_value_fx_portfolio",
    )

    currency_fx_rate = wb_filters.NumberFilter(label="FX Rates")
    currency_fx_rate__gte = wb_filters.NumberFilter(lookup_expr="gte", field_name="currency_fx_rate", label="FX Rates")
    currency_fx_rate__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="FX Rates",
        field_name="currency_fx_rate",
    )

    shares = wb_filters.NumberFilter(label="Shares")
    shares__gte = wb_filters.NumberFilter(lookup_expr="gte", field_name="shares", label="Shares")
    shares__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="Shares",
        field_name="shares",
    )

    def filter_classification(self, queryset, name, value):
        if value:
            return queryset.filter(underlying_instrument__in=value.get_classified_instruments())
        return queryset

    def filter_portfolio_instrument(self, queryset, name, value):
        if value:
            return queryset.filter(
                Q(portfolio__product=value) | Q(portfolio__product_group=value) | Q(portfolio__index=value)
            ).distinct()
        return queryset

    def boolean_hide_empty_position(self, queryset, name, value):
        if value:
            return queryset.exclude(
                (Q(shares__isnull=False) & Q(shares=0)) | Q(weighting=0) & Q(weighting__isnull=False)
            )
        return queryset

    class Meta:
        model = AssetPosition
        fields = {
            "is_estimated": ["exact"],
            # "exchange": ["exact"],
            "weighting": ["gte", "exact", "lte"],
            "portfolio": ["exact"],
            "portfolio_created": ["exact"],
            "currency": ["exact"],
            "underlying_instrument": ["exact"],
            "underlying_instrument__isin": ["exact"],
            "underlying_instrument__ticker": ["exact"],
        }


class AssetPositionPortfolioFilter(AssetPositionFilter):
    portfolio_instrument = portfolio = None
    aggregate = wb_filters.BooleanFilter(initial=False, required=True, label="Aggregate", method="filter_aggregate")

    def filter_aggregate(self, queryset, name, value):
        if value:
            annotate_map = {
                "id": Max("id"),
                "is_estimated": Max(Cast("is_estimated", output_field=IntegerField())),
                "asset_valuation_date": Max("asset_valuation_date"),
                "price": Max("price"),
                "shares": Sum("shares"),
                "total_value": Sum("total_value"),
                "currency_fx_rate": Max("currency_fx_rate"),
                "total_value_fx_portfolio": Sum("total_value_fx_portfolio"),
                "total_value_fx_usd": Sum("total_value_fx_usd"),
                "weighting": Sum("weighting"),
                "market_share": Max("market_share"),
                "liquidity": Max("liquidity"),
                "currency_symbol": Max("currency__symbol"),
                "underlying_quote_name": Max("underlying_quote__name"),
                "underlying_quote_ticker": Max("underlying_quote__ticker"),
                "underlying_quote_isin": Max("underlying_quote__isin"),
                "portfolio_currency_symbol": Max("portfolio__currency__symbol"),
            }
            for key in self.view._metric_serializer_fields.keys():
                annotate_map[key] = Max(key)

            return queryset.values(
                "underlying_instrument__name",
                "underlying_instrument__isin",
                "underlying_instrument__ticker",
            ).annotate(**annotate_map)
        return queryset

    class Meta:
        model = AssetPosition
        fields = {
            "is_estimated": ["exact"],
            "exchange": ["exact"],
            "weighting": ["gte", "exact", "lte"],
            "portfolio_created": ["exact"],
            "currency": ["exact"],
            "underlying_instrument": ["exact"],
        }


class AssetPositionInstrumentFilter(AssetPositionFilter):
    hide_empty_position = hide_aggregated_position = instrument_type = portfolio_instrument = is_cash = (
        classification
    ) = country = None
    date = wb_filters.DateFilter(
        label="Date",
        lookup_expr="exact",
        field_name="date",
        initial=get_latest_asset_position,
        method="filter_date",
        required=True,
    )
    filter_last_positions = wb_filters.BooleanFilter(label="Fetch last position", method="fake_filter")

    hide_aggregated_position = wb_filters.BooleanFilter(
        initial=False, method="boolean_hide_aggregated_position", label="Hide Position from Aggregated Product"
    )

    def boolean_hide_aggregated_position(self, queryset, name, value):
        if value:
            return queryset.exclude(is_invested=False)
        return queryset

    def filter_date(self, queryset, name, value):
        if value and self.request.GET.get("filter_last_positions", "false") == "false":
            return queryset.filter(date=value)
        return queryset

    class Meta:
        model = AssetPosition
        fields = {
            "exchange": ["exact"],
            "weighting": ["gte", "exact", "lte"],
            "portfolio": ["exact"],
            "portfolio_created": ["exact"],
        }


class AssetPositionProductGroupFilter(AssetPositionInstrumentFilter):
    pass


class AssetPositionIndexGroupFilter(AssetPositionInstrumentFilter):
    pass


class AssetPositionEquityGroupFilter(AssetPositionInstrumentFilter):
    pass


class CashPositionPortfolioFilterSet(wb_filters.FilterSet):
    underlying_instrument__instrument_type = wb_filters.ModelChoiceFilter(
        label="Instrument Type",
        endpoint=InstrumentType.get_representation_endpoint(),
        value_key=InstrumentType.get_representation_value_key(),
        label_key=InstrumentType.get_representation_label_key(),
        queryset=InstrumentType.objects.all(),
        lookup_expr="exact",
        field_name="underlying_instrument__instrument_type",
    )

    date = wb_filters.DateFilter(
        label="Date", lookup_expr="exact", field_name="date", initial=get_latest_asset_position, required=True
    )

    class Meta:
        model = AssetPosition
        fields = {
            "portfolio": ["exact"],
            "currency": ["exact"],
        }
        df_fields = {
            "total_value_fx_usd__gte": wb_filters.NumberFilter(
                lookup_expr="gte", field_name="total_value_fx_usd", label="Total Value ($)"
            ),
            "total_value_fx_usd__lte": wb_filters.NumberFilter(
                lookup_expr="lte",
                label="Total Value ($)",
                field_name="total_value_fx_usd",
            ),
        }


class DistributionFilter(wb_filters.FilterSet):
    date = wb_filters.DateFilter(
        label="Date",
        lookup_expr="exact",
        field_name="date",
        initial=latest_portfolio_date_based_on_instrument,
        required=True,
    )

    group_by = wb_filters.ChoiceFilter(
        label="Group By",
        choices=AssetPositionGroupBy.choices(),
        initial=AssetPositionGroupBy.INDUSTRY.name,
        method="fake_filter",
        clearable=False,
        required=True,
    )

    portfolio = wb_filters.ModelChoiceFilter(
        label="Portfolio",
        queryset=Portfolio.objects.all(),
        endpoint=Portfolio.get_representation_endpoint(),
        value_key=Portfolio.get_representation_value_key(),
        label_key=Portfolio.get_representation_label_key(),
        filter_params=get_portfolio_filter_params,
        initial=get_portfolio_default,
        required=True,
        clearable=False,
    )

    group_by_classification_group = wb_filters.ModelChoiceFilter(
        initial=lambda k, v, f: get_default_classification_group().id,
        method=lambda queryset, label, value: queryset,
        label="Group by Classification Group",
        queryset=ClassificationGroup.objects.all(),
        endpoint=ClassificationGroup.get_representation_endpoint(),
        value_key=ClassificationGroup.get_representation_value_key(),
        label_key=ClassificationGroup.get_representation_label_key(),
    )

    class Meta:
        model = AssetPosition
        fields = {}


def get_default_hedged_currency(field, request, view):
    with suppress(AttributeError, Portfolio.DoesNotExist, KeyError):
        return Portfolio.objects.get(id=view.kwargs["portfolio_id"]).hedged_currency.id


class ContributionChartFilter(wb_filters.FilterSet):
    date = wb_filters.FinancialPerformanceDateRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        required=True,
        clearable=False,
        initial=current_financial_month,
    )
    hedged_currency = wb_filters.ModelChoiceFilter(
        label="Hedged Currency",
        queryset=Currency.objects.all(),
        endpoint=Currency.get_representation_endpoint(),
        value_key=Currency.get_representation_value_key(),
        label_key=Currency.get_representation_label_key(),
        method=lambda queryset, label, value: queryset,
        initial=get_default_hedged_currency,
    )

    class Meta:
        model = AssetPosition
        fields = {}


class CompositionContributionChartFilter(ContributionChartFilter):
    show_lookthrough = wb_filters.BooleanFilter(
        label="Show Lookthrough", initial=False, required=True, method="fake_filter"
    )


class AssetPositionUnderlyingInstrumentChartFilter(DateFilterMixin, wb_filters.FilterSet):
    portfolio = wb_filters.ModelMultipleChoiceFilter(
        label="Portfolios",
        queryset=Portfolio.objects.all(),
        endpoint=Portfolio.get_representation_endpoint(),
        value_key=Portfolio.get_representation_value_key(),
        label_key=Portfolio.get_representation_label_key(),
    )

    class Meta:
        model = AssetPosition
        fields = {"underlying_instrument": ["exact"]}


class CompositionModelPortfolioPandasFilter(PandasFilterSetMixin, wb_filters.FilterSet):
    date = wb_filters.DateFilter(label="Date", lookup_expr="exact", initial=get_latest_asset_position, required=True)

    class Meta:
        model = AssetPosition
        fields = {}



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

VALUATIONDATAAUM_MENUITEM = MenuItem(
    label="AuM Chart",
    endpoint="wbportfolio:aumchart-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbportfolio.view_portfolio"]
    ),
)



from django.contrib import admin

from wbportfolio.models import Rebalancer, RebalancingModel


@admin.register(RebalancingModel)
class RebalancingModelAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = (
        "name",
        "class_path",
    )


@admin.register(Rebalancer)
class RebalancerAdmin(admin.ModelAdmin):
    search_fields = ("rebalancing_model__name",)
    list_display = (
        "computed_str",
        "rebalancing_model",
        "portfolio",
        "parameters",
        "approve_trade_proposal_automatically",
        "activation_date",
        "frequency",
    )



import pytest
from faker import Faker
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from wbcore.contrib.authentication.factories.users import UserFactory
from wbcore.contrib.directory.factories import EntryFactory
from wbcore.contrib.directory.models import Entry
from wbcore.utils.strings import format_number
from wbcrm.factories.accounts import AccountFactory, AccountRoleFactory
from wbcrm.models import Account

from wbportfolio.factories.claim import ClaimFactory
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.viewsets.transactions.claim import (
    ClaimAccountModelViewSet,
    ClaimAPIModelViewSet,
    ClaimEntryModelViewSet,
    ClaimModelViewSet,
    ClaimProductModelViewSet,
    ClaimRepresentationViewSet,
    ClaimTradeModelViewSet,
    ConsolidatedTradeSummaryDistributionChart,
    ConsolidatedTradeSummaryTableView,
    CumulativeNNMChartView,
    NegativeTermimalAccountPerProductModelViewSet,
    ProfitAndLossPandasView,
)

fake = Faker()


@pytest.mark.django_db
class TestClaimModelViewSet:
    def test_aggregate(self, claim):
        view = ClaimModelViewSet()
        queryset = Claim.objects.all()
        agg = view.get_aggregates(queryset, paginated_queryset=queryset)
        assert agg["shares"]["Σ"] == format_number(claim.shares, decimal=4)

    @pytest.fixture
    def account_user(self):
        # True, we create a superuser

        # if fake.pybool():
        #     user = UserFactory.create(is_superuser=True)
        # else:
        user = UserFactory.create(is_superuser=False)
        entry = Entry.objects.get(id=user.profile.id)

        # Create a bunch of account and roles
        public_account = AccountFactory.create(is_public=True, owner=EntryFactory.create())
        child_public_account = AccountFactory.create(parent=public_account, owner=EntryFactory.create())
        ClaimFactory.create(account=public_account)
        ClaimFactory.create(account=child_public_account)
        AccountRoleFactory.create(account=public_account, entry=entry)
        AccountRoleFactory.create(account=public_account)
        private_account = AccountFactory.create(is_public=False, owner=EntryFactory.create())
        child_private_account = AccountFactory.create(parent=private_account, owner=EntryFactory.create())
        ClaimFactory.create(account=private_account)
        ClaimFactory.create(account=child_private_account)
        return user

    @pytest.mark.parametrize(
        "viewset_class",
        [
            ClaimAPIModelViewSet,
            ClaimRepresentationViewSet,
            ClaimModelViewSet,
            ClaimAccountModelViewSet,
            ClaimProductModelViewSet,
            ClaimEntryModelViewSet,
            ClaimTradeModelViewSet,
            ConsolidatedTradeSummaryTableView,
            ConsolidatedTradeSummaryDistributionChart,
            CumulativeNNMChartView,
            ProfitAndLossPandasView,
        ],
    )
    def test_ensure_permission_on_account(self, account_user, viewset_class):
        """
        We ensure that all claims viewset doesn't show more that what the user is allowed to see.
        For claim, the allowed claim are all the claims where the account is among the account they is allowed to see
        """
        allowed_accounts = Account.objects.filter_for_user(account_user)
        allowed_claims = Claim.objects.filter(account__in=allowed_accounts)

        request = APIRequestFactory().get("")
        request.user = account_user
        viewset = viewset_class(
            request=request,
            kwargs={
                "entry_id": allowed_accounts.first().owner.id,
                "account_id": allowed_accounts.filter(parent__isnull=False).first().id,
                "product_id": allowed_claims.first().product.id,
                "trade_id": allowed_claims.first().trade.id,
            },
        )
        assert allowed_claims.exists()
        assert allowed_claims.count() < Claim.objects.count()  # Ensure that the filtering works
        for claim in viewset.get_queryset():
            assert claim in allowed_claims

    def test_ensure_permission_on_terminal_account_negative_sum_view(
        self, user, negative_claim_factory, account_with_owner_factory
    ):
        entry = Entry.objects.get(id=user.profile.id)

        request = APIRequestFactory().get("")
        request.user = user
        viewset = NegativeTermimalAccountPerProductModelViewSet(request=request)
        # Create a bunch of account and roles
        public_account = account_with_owner_factory.create(is_public=True)
        AccountRoleFactory.create(account=public_account, entry=entry)

        # create two negative claim, one only for an account the user can see
        private_account = account_with_owner_factory.create(is_public=False)
        public_claim = negative_claim_factory.create(account=public_account)  # noqa
        private_claim = negative_claim_factory.create(account=private_account)  # noqa

        assert set(viewset.get_queryset().values_list("account_id", flat=True)) == {public_account.id}

    def test_claim_validate_product_valuations_exist(
        self, claim_factory, super_user, weekday, instrument_price_factory
    ):
        claim = claim_factory.create(create_product_val=False, date=weekday)
        client = APIClient()
        url = reverse("wbportfolio:claim-submit", args=[claim.id])
        client.force_authenticate(user=super_user)
        response = client.patch(url)
        assert response.status_code == 412
        # check that the response contains "date", which means the error was detected on the "date" field
        assert response.json()["date"]

        # test that estimated price don't count
        instrument_price_factory.create(instrument=claim.product, date=weekday, calculated=True)
        response = client.patch(url)
        assert response.status_code == 412

        # With a valid price we can now submit
        instrument_price_factory.create(instrument=claim.product, date=weekday, calculated=False)
        response = client.patch(url)
        assert response.status_code == 200

        url = reverse("wbportfolio:claim-approve", args=[claim.id])
        client.force_authenticate(user=super_user)
        response = client.patch(url)
        assert response.status_code == 200



# Generated by Django 5.0.13 on 2025-03-21 12:30

import django.db.models.expressions
from decimal import Decimal
from django.db import migrations, models

def migrate_default_values(apps, schema_editor):
    Transaction = apps.get_model("wbportfolio", "Transaction")
    Trade = apps.get_model("wbportfolio", "Trade")
    Expiry = apps.get_model("wbportfolio", "Expiry")
    DividendTransaction = apps.get_model("wbportfolio", "DividendTransaction")
    Transaction.objects.filter(book_date__isnull=True).update(book_date=models.F("transaction_date"))
    Transaction.objects.filter(value_date__isnull=True).update(value_date=models.F("transaction_date"))
    Trade.objects.filter(price_gross__isnull=True).update(price_gross=models.F("price"))
    Expiry.objects.filter(price_gross__isnull=True).update(price_gross=models.F("price"))
    DividendTransaction.objects.filter(price_gross__isnull=True).update(price_gross=models.F("price"))
    Transaction.objects.filter(total_value_gross__isnull=True).update(total_value_gross=models.F("total_value"))


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0075_portfolio_initial_position_date_and_more'),
    ]

    operations = [
        migrations.RunPython(migrate_default_values),
        migrations.AlterField(
            model_name='transaction',
            name='book_date',
            field=models.DateField(help_text='The date that this transaction was booked.',
                                   verbose_name='Trade Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='transaction',
            name='value_date',
            field=models.DateField(help_text='The date that this transaction was valuated.',
                                   verbose_name='Value Date'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='price',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The price per share.', max_digits=16, verbose_name='Price'),
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='price_gross',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The gross price per share.', max_digits=16, verbose_name='Gross Price'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='dividendtransaction',
            name='shares',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The number of shares that were traded.', max_digits=15, verbose_name='Shares'),
        ),
        migrations.AlterField(
            model_name='expiry',
            name='price',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The price per share.', max_digits=16, verbose_name='Price'),
        ),
        migrations.AlterField(
            model_name='expiry',
            name='price_gross',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The gross price per share.', max_digits=16, verbose_name='Gross Price'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='expiry',
            name='shares',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The number of shares that were traded.', max_digits=15, verbose_name='Shares'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='price',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The price per share.', max_digits=16, verbose_name='Price'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='price_gross',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The gross price per share.', max_digits=16, verbose_name='Gross Price'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='trade',
            name='shares',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The number of shares that were traded.', max_digits=15, verbose_name='Shares'),
        ),

        migrations.AlterField(
            model_name='transaction',
            name='total_value',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), max_digits=20, verbose_name='Total Value'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='transaction',
            name='total_value_gross',
            field=models.DecimalField(decimal_places=4, default=Decimal('0.0'), max_digits=20, verbose_name='Total Value Gross'),
            preserve_default=False,
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='total_value_fx_portfolio',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='total_value_gross_fx_portfolio',
        ),
        migrations.AddField(
            model_name='transaction',
            name='total_value_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(
                models.F('currency_fx_rate'), '*', models.F('total_value')),
                                        output_field=models.DecimalField(decimal_places=4, max_digits=20)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='transaction',
            name='total_value_gross_fx_portfolio',
            field=models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*', models.F('total_value_gross')), output_field=models.DecimalField(decimal_places=4, max_digits=20)),
            preserve_default=False,
        ),

    ]



from import_export import fields
from import_export.widgets import DecimalWidget, ForeignKeyWidget
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.io.resources import ViewResource
from wbfdm.models import Exchange

from wbportfolio.models import AssetPosition


class AssetPositionResource(ViewResource):
    """
    The resource to download AssetPositions
    """

    exchange = fields.Field(
        column_name="exchange",
        attribute="exchange",
        widget=ForeignKeyWidget(Exchange, field="name"),
    )

    price = fields.Field(column_name="price", attribute="price", widget=DecimalWidget())
    shares = fields.Field(column_name="shares", attribute="shares", widget=DecimalWidget())
    total_value = fields.Field(column_name="total_value", attribute="total_value", widget=DecimalWidget())
    currency_fx_rate = fields.Field(
        column_name="currency_fx_rate", attribute="currency_fx_rate", widget=DecimalWidget()
    )
    total_value_fx_portfolio = fields.Field(
        column_name="total_value_fx_portfolio", attribute="total_value_fx_portfolio", widget=DecimalWidget()
    )
    total_value_fx_usd = fields.Field(
        column_name="total_value_fx_portfolio", attribute="total_value_fx_portfolio", widget=DecimalWidget()
    )

    currency = fields.Field(
        column_name="currency",
        attribute="currency",
        widget=ForeignKeyWidget(Currency, field="key"),
    )

    class Meta:
        import_id_fields = ("id",)
        fields = (
            "date",
            "exchange",
            "price",
            "shares",
            "total_value",
            "currency",
            "currency_fx_rate",
            "total_value_fx_portfolio",
            "total_value_fx_usd",
            "weighting",
        )
        export_order = (
            "date",
            "underlying_quote_name",
            "underlying_quote_ticker",
            "underlying_quote_isin",
            "exchange",
            "price",
            "shares",
            "total_value",
            "currency",
            "currency_fx_rate",
            "total_value_fx_portfolio",
            "total_value_fx_usd",
            "weighting",
        )
        model = AssetPosition
